# Generated by Django 2.0.5 on 2018-05-11 13:30

import django.contrib.postgres.fields.jsonb
import django.core.serializers.json
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("elasticsearch_django", "0006_add_encoder_JSONField_kwarg")]

    operations = [
        migrations.AlterField(
            model_name="searchquery",
            name="hits",
            field=django.contrib.postgres.fields.jsonb.JSONField(
                encoder=django.core.serializers.json.DjangoJSONEncoder,
                help_text="The list of meta info for each of the query matches returned.",
            ),
        ),
        migrations.AlterField(
            model_name="searchquery",
            name="query",
            field=django.contrib.postgres.fields.jsonb.JSONField(
                encoder=django.core.serializers.json.DjangoJSONEncoder,
                help_text="The raw ElasticSearch DSL query.",
            ),
        ),
    ]



from unittest import mock

from django.core.exceptions import ImproperlyConfigured
from django.test import TestCase

from ..apps import (
    ElasticAppConfig,
    _validate_config,
    _validate_model,
    _validate_mapping,
    _connect_signals,
    _on_model_delete,
    _on_model_save,
    _update_search_index,
)
from .. import tests
from ..models import SearchDocumentMixin
from ..tests import TestModel


class SearchAppsConfigTests(TestCase):

    """Tests for the apps module ready function."""

    @mock.patch("elasticsearch_django.apps.settings.get_setting")
    @mock.patch("elasticsearch_django.apps.settings.get_settings")
    @mock.patch("elasticsearch_django.apps._validate_config")
    @mock.patch("elasticsearch_django.apps._connect_signals")
    def test_ready(self, mock_signals, mock_config, mock_settings, mock_setting):
        """Test the AppConfig.ready method."""
        mock_setting.return_value = True  # auto-sync
        config = ElasticAppConfig("foo_bar", tests)
        config.ready()
        mock_config.assert_called_once_with(mock_setting.return_value)
        mock_signals.assert_called_once_with()

        mock_setting.return_value = False  # auto-sync
        mock_signals.reset_mock()
        mock_config.reset_mock()
        config.ready()
        mock_config.assert_called_once_with(mock_setting.return_value)
        mock_signals.assert_not_called()


class SearchAppsValidationTests(TestCase):

    """Tests for the apps module validation functions."""

    def test__validate_model(self):
        """Test _validate_model function."""
        # 1. model doesn't implement as_search_document
        with mock.patch("elasticsearch_django.tests.test_apps.TestModel") as tm:
            del tm.as_search_document
            self.assertRaises(ImproperlyConfigured, _validate_model, tm)

        # 2. model.objects doesn't implement get_search_queryset
        with mock.patch("elasticsearch_django.tests.test_apps.TestModel") as tm:
            del tm.objects.get_search_queryset
            self.assertRaises(ImproperlyConfigured, _validate_model, tm)

        # model should pass
        with mock.patch("elasticsearch_django.tests.test_apps.TestModel") as tm:
            _validate_model(tm)

    @mock.patch("elasticsearch_django.apps.settings")
    def test__validate_mapping(self, mock_settings):
        """Test _validate_model function."""
        _validate_mapping("foo", strict=True)
        mock_settings.get_index_mapping.assert_called_once_with("foo")
        mock_settings.get_index_mapping.side_effect = IOError()
        self.assertRaises(ImproperlyConfigured, _validate_mapping, "foo", strict=True)
        # shouldn't raise error
        _validate_mapping("foo", strict=False)

    @mock.patch("elasticsearch_django.apps.settings")
    @mock.patch("elasticsearch_django.apps._validate_model")
    @mock.patch("elasticsearch_django.apps._validate_mapping")
    def test__validate_config(self, mock_mapping, mock_model, mock_settings):
        """Test _validate_model function."""
        mock_settings.get_index_names.return_value = ["foo"]
        mock_settings.get_setting.return_value = "full"
        mock_settings.get_index_models.return_value = [TestModel]
        _validate_config()
        mock_mapping.assert_called_once_with("foo", strict=False)
        mock_model.assert_called_once_with(TestModel)

    @mock.patch("elasticsearch_django.apps.settings")
    @mock.patch("elasticsearch_django.apps._validate_model")
    @mock.patch("elasticsearch_django.apps._validate_mapping")
    def test__validate_config_invalid_strategy(
        self, mock_mapping, mock_model, mock_settings
    ):
        """Test _validate_model function with an invalid update_strategy."""
        mock_settings.get_index_names.return_value = ["foo"]
        mock_settings.get_setting.return_value = "foo"
        mock_settings.get_index_models.return_value = [TestModel]
        self.assertRaises(ImproperlyConfigured, _validate_config)

    @mock.patch("elasticsearch_django.apps.signals")
    @mock.patch("elasticsearch_django.apps.settings")
    def test__connect_signals(self, mock_settings, mock_signals):
        """Test the _connect_signals function."""
        # this should connect up the signals once, for TestModel
        mock_settings.get_index_names.return_value = ["foo"]
        mock_settings.get_index_models.return_value = [TestModel]
        _connect_signals()
        mock_signals.post_save.connect.assert_called_once_with(
            _on_model_save, sender=TestModel, dispatch_uid="testmodel.post_save"
        )
        mock_signals.post_delete.connect.assert_called_once_with(
            _on_model_delete, sender=TestModel, dispatch_uid="testmodel.post_delete"
        )

    @mock.patch.object(SearchDocumentMixin, "search_indexes", ["foo"])
    @mock.patch.object(SearchDocumentMixin, "delete_search_document")
    def test__on_model_delete(self, mock_delete):
        """Test the _on_model_delete function."""
        obj = SearchDocumentMixin()
        _on_model_delete(None, instance=obj)
        mock_delete.assert_called_once_with(index="foo")

    @mock.patch.object(SearchDocumentMixin, "search_indexes", ["foo", "bar"])
    @mock.patch.object(SearchDocumentMixin, "delete_search_document")
    def test__on_model_delete__multiple_indexes(self, mock_delete):
        """Test the _on_model_delete function with multiple indexes."""
        obj = SearchDocumentMixin()
        _on_model_delete(None, instance=obj)
        self.assertEqual(mock_delete.call_count, 2)

    @mock.patch("elasticsearch_django.apps._update_search_index")
    def test__on_model_save__index(self, mock_update):
        """Test the _on_model_save function without update_fields."""
        obj = mock.Mock(spec=SearchDocumentMixin, search_indexes=["foo"])
        _on_model_save(None, instance=obj, update_fields=None)
        mock_update.assert_called_once_with(
            instance=obj, index="foo", update_fields=None
        )

    @mock.patch("elasticsearch_django.apps._update_search_index")
    def test__on_model_save__update(self, mock_update):
        """Test the _on_model_save function without update_fields."""
        obj = mock.Mock(spec=SearchDocumentMixin, search_indexes=["foo"])
        _on_model_save(None, instance=obj, update_fields=["bar"])
        mock_update.assert_called_once_with(
            instance=obj, index="foo", update_fields=["bar"]
        )

    def test__update_search_index(self):
        """Test the _update_search_index function with an index action."""
        obj = mock.Mock(spec=SearchDocumentMixin)
        _update_search_index(instance=obj, index="foo", update_fields=None)
        self.assertEqual(obj.index_search_document.call_count, 1)
        self.assertEqual(obj.update_search_document.call_count, 0)
        obj.index_search_document.assert_called_once_with(index="foo")

    def test__update_search_update(self):
        """Test the _update_search_index function with an update action."""
        obj = mock.Mock(spec=SearchDocumentMixin)
        _update_search_index(instance=obj, index="foo", update_fields=["bar"])
        self.assertEqual(obj.update_search_document.call_count, 1)
        self.assertEqual(obj.index_search_document.call_count, 0)
        obj.update_search_document.assert_called_once_with(
            index="foo", update_fields=["bar"]
        )



# Generated by Django 1.9 on 2016-11-29 11:35

from django.db import migrations

from ..db.fields import JSONField


class Migration(migrations.Migration):

    dependencies = [("elasticsearch_django", "0003_auto_20160926_2021")]

    operations = [
        migrations.AlterField(
            model_name="searchquery",
            name="hits",
            field=JSONField(
                help_text="The list of meta info for each of the query matches returned."
            ),
        ),
        migrations.AlterField(
            model_name="searchquery",
            name="query",
            field=JSONField(help_text="The raw ElasticSearch DSL query."),
        ),
    ]



from unittest import mock

from django.test import TestCase
from django.db.models import signals
from elasticsearch_django.apps import _on_model_save
from elasticsearch_django.decorators import disable_search_updates, _strip_on_model_save

from . import TestModel, TestModelManager, SEARCH_DOC


class DecoratorTests(TestCase):

    """elasticsearch_django.decorators tests."""

    def setUp(self):
        signals.post_save.connect(_on_model_save)

    def tearDown(self):
        signals.post_save.disconnect(_on_model_save)

    def test_disable_updates(self):
        """Check the decorator removes _on_model_save from signal receivers."""
        self.assertEqual(len(signals.post_save.receivers), 1)
        self.assertEqual(signals.post_save.receivers[0][1](), _on_model_save)
        with disable_search_updates():
            self.assertEqual(signals.post_save.receivers, [])
        self.assertEqual(len(signals.post_save.receivers), 1)
        self.assertEqual(signals.post_save.receivers[0][1](), _on_model_save)

    def test__strip_on_model_save(self):
        self.assertEqual(len(signals.post_save.receivers), 1)
        self.assertEqual(_strip_on_model_save(), [])



import logging
import time
import warnings

from django.conf import settings
from django.contrib.postgres.fields import JSONField
from django.core.cache import cache
from django.core.serializers.json import DjangoJSONEncoder
from django.db import models
from django.db.models.expressions import RawSQL
from django.db.models.fields import CharField
from django.utils.timezone import now as tz_now

from .settings import (
    get_client,
    get_setting,
    get_model_indexes,
    get_model_index_properties,
)

logger = logging.getLogger(__name__)

UPDATE_STRATEGY_FULL = "full"
UPDATE_STRATEGY_PARTIAL = "partial"
UPDATE_STRATEGY = get_setting("update_strategy", UPDATE_STRATEGY_FULL)


class InvalidUpdateFields(Exception):
    """Custom exception raised when passing model fields to the
    save method as 'update_fields' that cannot be simply serialized."""

    def __init__(self, invalid_fields):
        message = (
            f"Unserializable update_fields value - please override "
            f"as_search_document_update: {invalid_fields}"
        )
        super().__init__(message)


class SearchDocumentManagerMixin(object):

    """
    Model manager mixin that adds search document methods.

    There is one method in this class that must implemented -
    `get_search_queryset`. This must return a queryset that is the
    set of objects to be indexed. This queryset is then converted
    into a generator that emits the objects as JSON documents.

    """

    def get_search_queryset(self, index="_all"):
        """
        Return the dataset used to populate the search index.

        Kwargs:
            index: string, the name of the index we are interested in -
                this allows us to have different sets of objects in
                different indexes. Defaults to '_all', in which case
                all indexes index the same set of objects.

        This must return a queryset object.

        """
        raise NotImplementedError(
            "{} does not implement 'get_search_queryset'.".format(
                self.__class__.__name__
            )
        )

    def in_search_queryset(self, instance_id, index="_all"):
        """
        Return True if an object is part of the search index queryset.

        Sometimes it's useful to know if an object _should_ be indexed. If
        an object is saved, how do you know if you should push that change
        to the search index? The simplest (albeit not most efficient) way
        is to check if it appears in the underlying search queryset.

        NB this method doesn't evaluate the entire dataset, it chains an
        additional queryset filter expression on the end. That's why it's
        important that the `get_search_queryset` method returns a queryset.

        Args:
            instance_id: the id of model object that we are looking for.

        Kwargs:
            index: string, the name of the index in which to check.
                Defaults to '_all'.

        """
        return self.get_search_queryset(index=index).filter(pk=instance_id).exists()

    def from_search_query(self, search_query):
        """
        Return queryset of objects from SearchQuery.results, **in order**.

        EXPERIMENTAL: this will only work with results from a single index,
        with a single doc_type - as we are returning a single QuerySet.

        This method takes the hits JSON and converts that into a queryset
        of all the relevant objects. The key part of this is the ordering -
        the order in which search results are returned is based on relevance,
        something that only ES can calculate, and that cannot be replicated
        in the database.

        It does this by adding custom SQL which annotates each record with
        the score from the search 'hit'. This is brittle, caveat emptor.

        The RawSQL clause is in the form:

            SELECT CASE {{model}}.id WHEN {{id}} THEN {{score}} END

        The "WHEN x THEN y" is repeated for every hit. The resulting SQL, in
        full is like this:

            SELECT "freelancer_freelancerprofile"."id",
                (SELECT CASE freelancer_freelancerprofile.id
                    WHEN 25 THEN 1.0
                    WHEN 26 THEN 1.0
                    [...]
                    ELSE 0
                END) AS "search_score"
            FROM "freelancer_freelancerprofile"
            WHERE "freelancer_freelancerprofile"."id" IN (25, 26, [...])
            ORDER BY "search_score" DESC

        It should be very fast, as there is no table lookup, but there is an
        assumption at the heart of this, which is that the search query doesn't
        contain the entire database - i.e. that it has been paged. (ES itself
        caps the results at 10,000.)

        """
        hits = search_query.hits
        score_sql = self._raw_sql([(h["id"], h["score"] or 0) for h in hits])
        rank_sql = self._raw_sql([(hits[i]["id"], i) for i in range(len(hits))])
        return (
            self.get_queryset()
            .filter(pk__in=[h["id"] for h in hits])
            # add the query relevance score
            .annotate(search_score=RawSQL(score_sql, ()))
            # add the ordering number (0-based)
            .annotate(search_rank=RawSQL(rank_sql, ()))
            .order_by("search_rank")
        )

    def _when(self, x, y):
        return "WHEN {} THEN {}".format(x, y)

    def _raw_sql(self, values):
        """Prepare SQL statement consisting of a sequence of WHEN .. THEN statements."""
        if isinstance(self.model._meta.pk, CharField):
            when_clauses = " ".join(
                [self._when("'{}'".format(x), y) for (x, y) in values]
            )
        else:
            when_clauses = " ".join([self._when(x, y) for (x, y) in values])
        table_name = self.model._meta.db_table
        primary_key = self.model._meta.pk.column
        return 'SELECT CASE {}."{}" {} ELSE 0 END'.format(
            table_name, primary_key, when_clauses
        )


class SearchDocumentMixin(object):

    """
    Mixin used by models that are indexed for ES.

    This mixin defines the interface exposed by models that
    are indexed ready for ES. The only method that needs
    implementing is `as_search_document`.

    """

    # Django model field types that can be serialized directly into
    # a known format. All other types will need custom serialization.
    # Used by as_search_document_update method
    SIMPLE_UPDATE_FIELD_TYPES = [
        "AutoField",
        "BooleanField",
        "CharField",
        "DateField",
        "DateTimeField",
        "DecimalField",
        "EmailField",
        "FloatField",
        "IntegerField",
        "TextField",
        "URLField",
        "UUIDField",
    ]

    @property
    def search_indexes(self):
        """Return the list of indexes for which this model is configured."""
        return get_model_indexes(self.__class__)

    @property
    def search_document_cache_key(self):
        """Key used for storing search docs in local cache."""
        return "elasticsearch_django:{}.{}.{}".format(
            self._meta.app_label, self._meta.model_name, self.pk
        )

    @property
    def search_doc_type(self):
        """Return the doc_type used for the model."""
        return self._meta.model_name

    def as_search_document(self, *, index):
        """
        Return the object as represented in a named index.

        This is named to avoid confusion - if it was `get_search_document`,
        which would be the logical name, it would not be clear whether it
        referred to getting the local representation of the search document,
        or actually fetching it from the index.

        Kwargs:
            index: string, the name of the index in which the object is to
                appear - this allows different representations in different
                indexes. Defaults to '_all', in which case all indexes use
                the same search document structure.

        Returns a dictionary.

        """
        raise NotImplementedError(
            "{} does not implement 'as_search_document'.".format(
                self.__class__.__name__
            )
        )

    def _is_field_serializable(self, field_name):
        """Return True if the field can be serialized into a JSON doc."""
        return (
            self._meta.get_field(field_name).get_internal_type()
            in self.SIMPLE_UPDATE_FIELD_TYPES
        )

    def clean_update_fields(self, index, update_fields):
        """
        Clean the list of update_fields based on the index being updated.\

        If any field in the update_fields list is not in the set of properties
        defined by the index mapping for this model, then we ignore it. If
        a field _is_ in the mapping, but the underlying model field is a
        related object, and thereby not directly serializable, then this
        method will raise a ValueError.

        """
        search_fields = get_model_index_properties(self, index)
        clean_fields = [f for f in update_fields if f in search_fields]
        ignore = [f for f in update_fields if f not in search_fields]
        if ignore:
            logger.debug(
                "Ignoring fields from partial update: %s",
                [f for f in update_fields if f not in search_fields],
            )
        for f in clean_fields:
            if not self._is_field_serializable(f):
                raise ValueError(
                    "'%s' cannot be automatically serialized into a search document property. Please override as_search_document_update.",
                    f,
                )
        return clean_fields

    def as_search_document_update(self, *, index, update_fields):
        """
        Return a partial update document based on which fields have been updated.

        If an object is saved with the `update_fields` argument passed
        through, then it is assumed that this is a 'partial update'. In
        this scenario we need a {property: value} dictionary containing
        just the fields we want to update.

        This method handles two possible update strategies - 'full' or 'partial'.
        The default 'full' strategy simply returns the value of `as_search_document`
        - thereby replacing the entire document each time. The 'partial' strategy is
        more intelligent - it will determine whether the fields passed are in the
        search document mapping, and return a partial update document that contains
        only those that are. In addition, if any field that _is_ included cannot
        be automatically serialized (e.g. a RelatedField object), then this method
        will raise a ValueError. In this scenario, you should override this method
        in your subclass.

        >>> def as_search_document_update(self, index, update_fields):
        ...     if 'user' in update_fields:
        ...         update_fields.remove('user')
        ...         doc = super().as_search_document_update(index, update_fields)
        ...         doc['user'] = self.user.get_full_name()
        ...         return doc
        ...     return super().as_search_document_update(index, update_fields)

        You may also wish to subclass this method to perform field-specific logic
        - in this example if only the timestamp is being saved, then ignore the
        update if the timestamp is later than a certain time.

        >>> def as_search_document_update(self, index, update_fields):
        ...     if update_fields == ['timestamp']:
        ...         if self.timestamp > today():
        ...            return {}
        ...     return super().as_search_document_update(index, update_fields)

        """
        if UPDATE_STRATEGY == UPDATE_STRATEGY_FULL:
            return self.as_search_document(index=index)

        if UPDATE_STRATEGY == UPDATE_STRATEGY_PARTIAL:
            # in partial mode we update the intersection of update_fields and
            # properties found in the mapping file.
            return {
                k: getattr(self, k)
                for k in self.clean_update_fields(
                    index=index, update_fields=update_fields
                )
            }

    def as_search_action(self, *, index, action):
        """
        Return an object as represented in a bulk api operation.

        Bulk API operations have a very specific format. This function will
        call the standard `as_search_document` method on the object and then
        wrap that up in the correct format for the action specified.

        https://www.elastic.co/guide/en/elasticsearch/reference/current/docs-bulk.html

        Args:
            index: string, the name of the index in which the action is to
                be taken. Bulk operations are only every carried out on a single
                index at a time.
            action: string ['index' | 'update' | 'delete'] - this decides
                how the final document is formatted.

        Returns a dictionary.

        """
        if action not in ("index", "update", "delete"):
            raise ValueError("Action must be 'index', 'update' or 'delete'.")

        document = {
            "_index": index,
            "_type": self.search_doc_type,
            "_op_type": action,
            "_id": self.pk,
        }

        if action == "index":
            document["_source"] = self.as_search_document(index=index)
        elif action == "update":
            document["doc"] = self.as_search_document(index=index)
        return document

    def fetch_search_document(self, *, index):
        """Fetch the object's document from a search index by id."""
        assert self.pk, "Object must have a primary key before being indexed."
        client = get_client()
        return client.get(index=index, doc_type=self.search_doc_type, id=self.pk)

    def index_search_document(self, *, index):
        """
        Create or replace search document in named index.

        Checks the local cache to see if the document has changed,
        and if not aborts the update, else pushes to ES, and then
        resets the local cache. Cache timeout is set as CACHE_EXPIRY
        in the settings, and defaults to 60s.

        """
        cache_key = self.search_document_cache_key
        new_doc = self.as_search_document(index=index)
        cached_doc = cache.get(cache_key)
        if new_doc == cached_doc:
            logger.debug("Search document for %r is unchanged, ignoring update.", self)
            return []
        cache.set(cache_key, new_doc, timeout=get_setting("CACHE_EXPIRY", 60))
        get_client().index(
            index=index, doc_type=self.search_doc_type, body=new_doc, id=self.pk
        )

    def update_search_document(self, *, index, update_fields):
        """
        Partial update of a document in named index.

        Partial updates are invoked via a call to save the document
        with 'update_fields'. These fields are passed to the
        as_search_document method so that it can build a partial
        document. NB we don't just call as_search_document and then
        strip the fields _not_ in update_fields as we are trying
        to avoid possibly expensive operations in building the
        source document. The canonical example for this method
        is updating a single timestamp on a model - we don't want
        to have to walk the model relations and build a document
        in this case - we just want to push the timestamp.

        When POSTing a partial update the `as_search_document` doc
        must be passed to the `client.update` wrapped in a "doc" node,
        see: https://www.elastic.co/guide/en/elasticsearch/reference/current/docs-update.html

        """
        doc = self.as_search_document_update(index=index, update_fields=update_fields)
        if not doc:
            logger.debug("Ignoring object update as document is empty.")
            return

        get_client().update(
            index=index, doc_type=self.search_doc_type, body={"doc": doc}, id=self.pk
        )

    def delete_search_document(self, *, index):
        """Delete document from named index."""
        cache.delete(self.search_document_cache_key)
        get_client().delete(index=index, doc_type=self.search_doc_type, id=self.pk)


class SearchQuery(models.Model):

    """
    Model used to capture ES queries and responses.

    For low-traffic sites it's useful to be able to replay
    searches, and to track how a user filtered and searched.
    This model can be used to store a search query and meta
    information about the results (document type, id and score).

    >>> from elasticsearch_dsl import Search
    >>> search = Search(using=client)
    >>> sq = SearchQuery.execute(search).save()

    """

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        related_name="search_queries",
        blank=True,
        null=True,
        help_text="The user who made the search query (nullable).",
        on_delete=models.SET_NULL,
    )
    index = models.CharField(
        max_length=100,
        default="_all",
        help_text="The name of the ElasticSearch index(es) being queried.",
    )
    # The query property contains the raw DSL query, which can be arbitrarily complex - there
    # is no one way of mapping input text to the query itself. However, it's often helpful to
    # have the terms that the user themselves typed easily accessible without having to parse
    # JSON.
    search_terms = models.CharField(
        max_length=400,
        default="",
        blank=True,
        help_text="Free text search terms used in the query, stored for easy reference.",
    )
    query = JSONField(
        help_text="The raw ElasticSearch DSL query.", encoder=DjangoJSONEncoder
    )
    hits = JSONField(
        help_text="The list of meta info for each of the query matches returned.",
        encoder=DjangoJSONEncoder,
    )
    total_hits = models.IntegerField(
        default=0,
        help_text="Total number of matches found for the query (!= the hits returned).",
    )
    reference = models.CharField(
        max_length=100,
        default="",
        blank=True,
        help_text="Custom reference used to identify and group related searches.",
    )
    executed_at = models.DateTimeField(
        help_text="When the search was executed - set via execute() method."
    )
    duration = models.FloatField(
        help_text="Time taken to execute the search itself, in seconds."
    )

    class Meta:
        app_label = "elasticsearch_django"
        verbose_name = "Search query"
        verbose_name_plural = "Search queries"

    def __str__(self):
        return "Query (id={}) run against index '{}'".format(self.pk, self.index)

    def __repr__(self):
        return "<SearchQuery id={} user={} index='{}' total_hits={} >".format(
            self.pk, self.user, self.index, self.total_hits
        )

    @classmethod
    def execute(cls, search, search_terms="", user=None, reference=None, save=True):
        """Create a new SearchQuery instance and execute a search against ES."""
        warnings.warn(
            "Pending deprecation - please use `execute_search` function instead.",
            PendingDeprecationWarning,
        )
        return execute_search(
            search, search_terms=search_terms, user=user, reference=reference, save=save
        )

    def save(self, **kwargs):
        """Save and return the object (for chaining)."""
        if self.search_terms is None:
            self.search_terms = ""
        super().save(**kwargs)
        return self

    def _extract_set(self, _property):
        return (
            [] if self.hits is None else (list(set([h[_property] for h in self.hits])))
        )

    @property
    def doc_types(self):
        """List of doc_types extracted from hits."""
        return self._extract_set("doc_type")

    @property
    def max_score(self):
        """The max relevance score in the returned page."""
        return max(self._extract_set("score") or [0])

    @property
    def min_score(self):
        """The min relevance score in the returned page."""
        return min(self._extract_set("score") or [0])

    @property
    def object_ids(self):
        """List of model ids extracted from hits."""
        return self._extract_set("id")

    @property
    def page_slice(self):
        """Return the query from:size tuple (0-based)."""
        return (
            None
            if self.query is None
            else (self.query.get("from", 0), self.query.get("size", 10))
        )

    @property
    def page_from(self):
        """1-based index of the first hit in the returned page."""
        return 0 if self.page_size == 0 else self.page_slice[0] + 1

    @property
    def page_to(self):
        """1-based index of the last hit in the returned page."""
        return 0 if self.page_size == 0 else self.page_from + self.page_size - 1

    @property
    def page_size(self):
        """The number of hits returned in this specific page."""
        return 0 if self.hits is None else len(self.hits)


def execute_search(search, search_terms="", user=None, reference=None, save=True):
    """
    Create a new SearchQuery instance and execute a search against ES.

    Args:
        search: elasticsearch.search.Search object, that internally contains
            the connection and query; this is the query that is executed. All
            we are doing is logging the input and parsing the output.
        search_terms: raw end user search terms input - what they typed into the search
            box.
        user: Django User object, the person making the query - used for logging
            purposes. Can be null.
        reference: string, can be anything you like, used for identification,
            grouping purposes.
        save: bool, if True then save the new object immediately, can be
            overridden to False to prevent logging absolutely everything.
            Defaults to True

    """
    start = time.time()
    response = search.execute()
    duration = time.time() - start
    log = SearchQuery(
        user=user,
        search_terms=search_terms,
        index=", ".join(search._index or ["_all"])[:100],  # field length restriction
        query=search.to_dict(),
        hits=[h.meta.to_dict() for h in response.hits],
        total_hits=response.hits.total,
        reference=reference or "",
        executed_at=tz_now(),
        duration=duration,
    )
    log.response = response
    return log.save() if save else log



import datetime
import decimal
from unittest import mock

from django.core.cache import cache
from django.db.models import Model
from django.test import TestCase
from django.utils.timezone import now as tz_now
from elasticsearch_dsl.search import Search

from ..models import (
    InvalidUpdateFields,
    SearchDocumentMixin,
    SearchDocumentManagerMixin,
    SearchQuery,
    UPDATE_STRATEGY_FULL,
    UPDATE_STRATEGY_PARTIAL,
)
from ..tests import TestModel, TestModelManager, SEARCH_DOC


class SearchDocumentMixinTests(TestCase):

    """Tests for the SearchDocumentMixin."""

    @mock.patch("elasticsearch_django.models.get_model_indexes")
    def test_search_indexes(self, mock_indexes):
        """Test the search_indexes function."""
        mock_indexes.return_value = "foo"
        obj = TestModel()
        self.assertEqual(obj.search_indexes, "foo")
        mock_indexes.assert_called_once_with(TestModel)

    def test_as_search_document(self):
        """Test the as_search_document method."""
        obj = SearchDocumentMixin()
        self.assertRaises(NotImplementedError, obj.as_search_document, index="_all")

    def test__is_field_serializable(self):
        obj = TestModel()
        self.assertTrue(obj._is_field_serializable("simple_field_1"))
        self.assertTrue(obj._is_field_serializable("simple_field_1"))
        self.assertFalse(obj._is_field_serializable("complex_field"))

    @mock.patch("elasticsearch_django.models.get_model_index_properties")
    def test_clean_update_fields(self, mock_properties):
        """Test that only fields in the mapping file are cleaned."""
        obj = TestModel()
        mock_properties.return_value = ["simple_field_1", "complex_field"]
        self.assertEqual(
            obj.clean_update_fields(
                index="", update_fields=["simple_field_1", "simple_field_2"]
            ),
            ["simple_field_1"],
        )

    @mock.patch("elasticsearch_django.models.get_model_index_properties")
    def test_clean_update_fields_complex_object(self, mock_properties):
        """Test that unserializable fields raise a ValueError."""
        obj = TestModel()
        mock_properties.return_value = ["simple_field_1", "complex_field"]
        self.assertRaises(
            ValueError,
            obj.clean_update_fields,
            index="",
            update_fields=["simple_field_1", "complex_field"],
        )

    @mock.patch("elasticsearch_django.models.get_model_index_properties")
    def test_as_search_document_update_full(self, mock_properties):
        """Test the as_search_document_update method."""
        obj = TestModel(simple_field_1=1, simple_field_2="foo")
        mock_properties.return_value = ["simple_field_1"]
        with mock.patch(
            "elasticsearch_django.models.UPDATE_STRATEGY", UPDATE_STRATEGY_FULL
        ):
            self.assertEqual(
                obj.as_search_document_update(
                    index="_all", update_fields=["simple_field_1"]
                ),
                obj.as_search_document(index="_all"),
            )

    @mock.patch("elasticsearch_django.models.UPDATE_STRATEGY", UPDATE_STRATEGY_PARTIAL)
    @mock.patch("elasticsearch_django.models.get_model_index_properties")
    def test_as_search_document_update_partial(self, mock_properties):
        """Test the as_search_document_update method."""
        obj = TestModel(simple_field_1=1, simple_field_2="foo")
        mock_properties.return_value = ["simple_field_1", "simple_field_2"]
        self.assertEqual(
            obj.as_search_document_update(
                index="_all", update_fields=["simple_field_1", "simple_field_2"]
            ),
            {
                "simple_field_1": obj.simple_field_1,
                "simple_field_2": obj.simple_field_2,
            },
        )
        # remove simple_field_2 from the mapping - should no longer be included
        mock_properties.return_value = ["simple_field_1"]
        self.assertEqual(
            obj.as_search_document_update(
                index="_all", update_fields=["simple_field_1", "simple_field_2"]
            ),
            {"simple_field_1": obj.simple_field_1},
        )

    @mock.patch(
        "elasticsearch_django.settings.get_connection_string",
        lambda: "http://testserver",
    )
    @mock.patch("elasticsearch_django.models.get_client")
    def test_index_search_document(self, mock_client):
        """Test the index_search_document sets the cache."""
        obj = TestModel(pk=1)
        doc = obj.as_search_document(index="_all")
        key = obj.search_document_cache_key
        self.assertIsNone(cache.get(key))
        obj.index_search_document(index="_all")
        self.assertEqual(cache.get(key), doc)
        mock_client.return_value.index.assert_called_once_with(
            body=doc, doc_type="testmodel", id=1, index="_all"
        )

    @mock.patch(
        "elasticsearch_django.settings.get_connection_string",
        lambda: "http://testserver",
    )
    @mock.patch("elasticsearch_django.models.get_client")
    def test_index_search_document_cached(self, mock_client):
        """Test the index_search_document does not update if doc is a duplicate."""
        obj = TestModel(pk=1)
        doc = obj.as_search_document(index="_all")
        key = obj.search_document_cache_key
        cache.set(key, doc, timeout=1)
        self.assertEqual(cache.get(key), doc)
        obj.index_search_document(index="_all")
        self.assertEqual(mock_client.call_count, 0)

    @mock.patch(
        "elasticsearch_django.settings.get_connection_string",
        lambda: "http://testserver",
    )
    @mock.patch("elasticsearch_django.models.get_client")
    def test_update_search_document(self, mock_client):
        """Test the update_search_document wraps up doc correctly."""
        obj = TestModel(pk=1, simple_field_1=1)
        doc = obj.as_search_document_update(
            index="_all", update_fields=["simple_field_1"]
        )
        obj.update_search_document(index="_all", update_fields=["simple_field_1"])
        mock_client.return_value.update.assert_called_once_with(
            body={"doc": doc}, doc_type="testmodel", id=1, index="_all"
        )

    @mock.patch(
        "elasticsearch_django.settings.get_connection_string",
        lambda: "http://testserver",
    )
    @mock.patch("elasticsearch_django.models.get_client")
    def test_update_search_document_empty(self, mock_client):
        """Test the update_search_document ignores empty updates."""
        obj = TestModel(pk=1, simple_field_1=1)
        with mock.patch.object(TestModel, "as_search_document_update") as mock_update:
            mock_update.return_value = {}
            # this will return an empty dictionary as the partial update doc
            obj.update_search_document(index="_all", update_fields=[])
            mock_client.return_value.update.assert_not_called()

    @mock.patch(
        "elasticsearch_django.settings.get_connection_string",
        lambda: "http://testserver",
    )
    @mock.patch("elasticsearch_django.models.get_client")
    def test_delete_search_document(self, mock_client):
        """Test the delete_search_document clears the cache."""
        obj = TestModel(pk=1)
        doc = obj.as_search_document(index="_all")
        key = obj.search_document_cache_key
        cache.set(key, doc)
        self.assertIsNotNone(cache.get(key))
        obj.delete_search_document(index="_all")
        self.assertIsNone(cache.get(key))
        mock_client.return_value.delete.assert_called_once_with(
            doc_type="testmodel", id=1, index="_all"
        )

    def test_as_search_action(self):
        """Test the as_search_action method."""
        obj = TestModel()

        # invalid action 'bar'
        self.assertRaises(ValueError, obj.as_search_action, index="foo", action="bar")

        self.assertEqual(
            obj.as_search_action(index="foo", action="index"),
            {
                "_index": "foo",
                "_type": "testmodel",
                "_op_type": "index",
                "_id": None,
                "_source": SEARCH_DOC,
            },
        )

        self.assertEqual(
            obj.as_search_action(index="foo", action="update"),
            {
                "_index": "foo",
                "_type": "testmodel",
                "_op_type": "update",
                "_id": None,
                "doc": SEARCH_DOC,
            },
        )

        self.assertEqual(
            obj.as_search_action(index="foo", action="delete"),
            {"_index": "foo", "_type": "testmodel", "_op_type": "delete", "_id": None},
        )

    @mock.patch("elasticsearch_django.models.get_client")
    def test_fetch_search_document(self, mock_client):
        """Test the fetch_search_document method."""
        obj = TestModel()
        # obj has no id
        self.assertRaises(AssertionError, obj.fetch_search_document, index="foo")

        # should now call the ES get method
        obj.id = 1
        response = obj.fetch_search_document(index="foo")
        mock_get = mock_client.return_value.get
        mock_get.assert_called_once_with(
            index="foo", doc_type=obj._meta.model_name, id=obj.id
        )
        self.assertEqual(response, mock_get.return_value)


class SearchDocumentManagerMixinTests(TestCase):

    """Tests for the SearchDocumentManagerMixin."""

    def test_get_search_queryset(self):
        """Test the get_search_queryset method."""
        obj = SearchDocumentManagerMixin()
        self.assertRaises(NotImplementedError, obj.get_search_queryset)

    @mock.patch.object(TestModelManager, "get_search_queryset")
    def test_in_search_queryset(self, mock_qs):
        """Test the in_search_queryset method."""
        obj = TestModel(id=1)
        TestModel.objects.in_search_queryset(obj.id)
        mock_qs.assert_called_once_with(index="_all")
        mock_qs.return_value.filter.assert_called_once_with(pk=1)
        mock_qs.return_value.filter.return_value.exists.assert_called_once_with()

    def test__raw_sql(self):
        """Test the _raw_sql method."""
        self.assertEqual(
            TestModel.objects._raw_sql(((1, 2), (3, 4))),
            'SELECT CASE elasticsearch_django_testmodel."id" WHEN 1 THEN 2 WHEN 3 THEN 4 ELSE 0 END',
        )

    @mock.patch("django.db.models.query.QuerySet")
    def test_from_search_query(self, mock_qs):
        """Test the from_search_query method."""
        self.maxDiff = None
        sq = SearchQuery(hits=[{"id": 1, "score": 1}, {"id": 2, "score": 2}])
        qs = TestModel.objects.from_search_query(sq)
        self.assertEqual(
            str(qs.query),
            'SELECT "elasticsearch_django_testmodel"."id", "elasticsearch_django_testmodel"."simple_field_1", '  # noqa
            '"elasticsearch_django_testmodel"."simple_field_2", "elasticsearch_django_testmodel"."complex_field", '  # noqa
            '(SELECT CASE elasticsearch_django_testmodel."id" WHEN 1 THEN 1 WHEN 2 THEN 2 ELSE 0 END) AS "search_score", '  # noqa
            '(SELECT CASE elasticsearch_django_testmodel."id" WHEN 1 THEN 0 WHEN 2 THEN 1 ELSE 0 END) AS "search_rank" '  # noqa
            'FROM "elasticsearch_django_testmodel" WHERE "elasticsearch_django_testmodel"."id" IN (1, 2) ORDER BY "search_rank" ASC',  # noqa
        )

        # test with a null score - new in v5
        sq = SearchQuery(hits=[{"id": 1, "score": None}, {"id": 2, "score": 2}])
        qs = TestModel.objects.from_search_query(sq)
        self.assertEqual(
            str(qs.query),
            'SELECT "elasticsearch_django_testmodel"."id", "elasticsearch_django_testmodel"."simple_field_1", '  # noqa
            '"elasticsearch_django_testmodel"."simple_field_2", "elasticsearch_django_testmodel"."complex_field", '  # noqa
            '(SELECT CASE elasticsearch_django_testmodel."id" WHEN 1 THEN 0 WHEN 2 THEN 2 ELSE 0 END) AS "search_score", '  # noqa
            '(SELECT CASE elasticsearch_django_testmodel."id" WHEN 1 THEN 0 WHEN 2 THEN 1 ELSE 0 END) AS "search_rank" '  # noqa
            'FROM "elasticsearch_django_testmodel" WHERE "elasticsearch_django_testmodel"."id" IN (1, 2) ORDER BY "search_rank" ASC',  # noqa
        )


class SearchQueryTests(TestCase):

    """Tests for the SearchQuery model."""

    hits = [
        {"id": 1, "doc_type": "foo"},
        {"id": 2, "doc_type": "foo"},
        {"id": 3, "doc_type": "bar"},
    ]

    def test__extract_set(self):
        """Test the _extract_set method."""
        obj = SearchQuery(hits=SearchQueryTests.hits)
        self.assertEqual(set(obj._extract_set("id")), set([1, 2, 3]))

    def test_doc_types(self):
        """Test the doc_types property."""
        obj = SearchQuery(hits=SearchQueryTests.hits)
        self.assertEqual(set(obj.doc_types), set(["foo", "bar"]))

    def test_object_ids(self):
        """Test the object_ids property."""
        obj = SearchQuery(hits=SearchQueryTests.hits)
        self.assertEqual(set(obj.object_ids), set([1, 2, 3]))

    def test_save(self):
        """Try saving unserializable JSON."""
        today = datetime.date.today()
        sq = SearchQuery(
            user=None,
            index="foo",
            query={"today": today},
            hits={"hits": decimal.Decimal("1.0")},
            total_hits=100,
            reference="bar",
            executed_at=tz_now(),
            duration=0,
        )
        sq.save()
        sq.refresh_from_db()
        # invalid JSON values will have been converted
        self.assertEqual(sq.search_terms, "")
        self.assertEqual(sq.query["today"], today.isoformat())
        self.assertEqual(sq.hits["hits"], "1.0")

    def test_paging(self):
        """Test the paging properties."""
        sq = SearchQuery()
        self.assertEqual(sq.page_slice, None)

        # no hits, so should all be 0
        sq.query = {"from": 0, "size": 25}
        self.assertEqual(sq.page_slice, (0, 25))
        self.assertEqual(sq.page_from, 0)
        self.assertEqual(sq.page_to, 0)
        self.assertEqual(sq.page_size, 0)

        # three hits
        sq.hits = [1, 2, 3]  # random list of size = 3
        sq.query = {"from": 0, "size": 25}
        self.assertEqual(sq.page_from, 1)
        self.assertEqual(sq.page_to, 3)
        self.assertEqual(sq.page_size, 3)

    def test_scores(self):
        """Test the max/min properties."""
        sq = SearchQuery()
        self.assertEqual(sq.max_score, 0)
        self.assertEqual(sq.min_score, 0)

        sq.hits = [{"score": 1}, {"score": 2}]
        self.assertEqual(sq.max_score, 2)
        self.assertEqual(sq.min_score, 1)

    @mock.patch("elasticsearch_django.models.tz_now")
    @mock.patch.object(Search, "execute")
    @mock.patch.object(Model, "save")
    def test_execute(self, mock_save, mock_execute, mock_now):
        """Test the execute class method."""
        search = Search()
        sq = SearchQuery.execute(search)
        self.assertEqual(sq.user, None)
        self.assertEqual(sq.search_terms, "")
        self.assertEqual(sq.index, "_all")
        self.assertEqual(sq.query, search.to_dict())
        self.assertEqual(sq.hits, [])
        self.assertEqual(sq.total_hits, mock_execute.return_value.hits.total)
        self.assertEqual(sq.reference, "")
        self.assertTrue(sq.duration > 0)
        self.assertEqual(sq.executed_at, mock_now.return_value)
        self.assertEqual(sq.response, mock_execute.return_value)
        mock_save.assert_called_once_with()

        # try without saving
        mock_save.reset_mock()
        sq = SearchQuery.execute(search, save=False)
        mock_save.assert_not_called()



from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [("elasticsearch_django", "0001_initial")]

    operations = [
        migrations.AddField(
            model_name="searchquery",
            name="duration",
            field=models.FloatField(
                default=0,
                help_text="Time taken to execute the search itself, in seconds.",
            ),
            preserve_default=False,
        )
    ]



from unittest import mock

from django.test import TestCase
from django.test.utils import override_settings

from ..settings import (
    get_client,
    get_setting,
    get_settings,
    set_setting,
    get_connection_string,
    get_index_config,
    get_index_names,
    get_index_models,
    get_model_indexes,
    get_document_models,
    get_document_model,
    auto_sync,
)
from ..tests import TestModel


TEST_SETTINGS = {
    "connections": {"default": "https://foo", "backup": "https://bar"},
    "indexes": {"baz": {"models": ["elasticsearch_django.TestModel"]}},
    "settings": {"foo": "bar", "auto_sync": True, "never_auto_sync": []},
}


class SettingsFunctionTests(TestCase):

    """Tests for the settings functions."""

    @mock.patch("elasticsearch_django.settings.get_connection_string")
    def test_get_client(self, mock_conn):
        """Test the get_client function."""
        mock_conn.return_value = "http://foo"
        client = get_client()
        self.assertEqual(len(client.transport.hosts), 1)
        self.assertEqual(client.transport.hosts[0]["host"], "foo")

    @override_settings(SEARCH_SETTINGS=TEST_SETTINGS)
    def test_get_settings(self):
        """Test the get_settings method."""
        self.assertEqual(get_settings(), TEST_SETTINGS["settings"])

    @override_settings(SEARCH_SETTINGS=TEST_SETTINGS)
    def test_get_setting(self):
        """Test the get_setting method."""
        self.assertEqual(get_setting("foo"), "bar")

    @override_settings(SEARCH_SETTINGS=TEST_SETTINGS)
    def test_get_setting_with_default(self):
        """Test the get_setting method."""
        self.assertRaises(KeyError, get_setting, "bar")
        self.assertEqual(get_setting("bar", "baz"), "baz")

    @override_settings(SEARCH_SETTINGS=TEST_SETTINGS)
    def test_get_connection_string(self):
        """Test the get_connection_string method."""
        self.assertEqual(
            get_connection_string(), TEST_SETTINGS["connections"]["default"]
        )
        self.assertEqual(
            get_connection_string("backup"), TEST_SETTINGS["connections"]["backup"]
        )

    @override_settings(SEARCH_SETTINGS=TEST_SETTINGS)
    def test_get_index_config(self):
        """Test the get_index_config method."""
        self.assertEqual(get_index_config("baz"), TEST_SETTINGS["indexes"]["baz"])

    @override_settings(SEARCH_SETTINGS=TEST_SETTINGS)
    def test_get_index_names(self):
        """Test the get_index_names method."""
        self.assertEqual(get_index_names(), list(TEST_SETTINGS["indexes"].keys()))

    @override_settings(SEARCH_SETTINGS=TEST_SETTINGS)
    def test_get_index_models(self):
        """Test the get_index_models function."""
        from django.apps import apps

        models = get_index_models("baz")
        self.assertEqual(models, [apps.get_model("elasticsearch_django", "TestModel")])

    @override_settings(SEARCH_SETTINGS=TEST_SETTINGS)
    def test_get_model_indexes(self):
        """Test the get_model_indexes function."""
        # TestModel is in the TEST_SETTINGS
        self.assertEqual(get_model_indexes(TestModel), ["baz"])
        # plain old object isn't in any indexes
        self.assertEqual(get_model_indexes(object), [])

    def test_get_index_mapping(self):
        """Test the get_index_mapping function."""
        # this interacts with the file system, not going to bother to test
        # as it just opens a file and loads in into a dict - there's no 'logic'
        pass

    @override_settings(SEARCH_SETTINGS=TEST_SETTINGS)
    def test_get_document_models(self):
        """Test the get_document_models function."""
        self.assertEqual(get_document_models(), {"baz.testmodel": TestModel})

    @override_settings(SEARCH_SETTINGS=TEST_SETTINGS)
    def test_get_document_model(self):
        """Test the get_document_model function."""
        self.assertEqual(get_document_model("baz", "testmodel"), TestModel)

    @override_settings(SEARCH_SETTINGS=TEST_SETTINGS)
    def test_auto_sync(self):
        """Test the auto_sync function."""
        obj = TestModel()
        self.assertEqual(auto_sync(obj), True)
        # Check that if the setting auto_sync is False, the function auto_sync also returns false.
        TEST_SETTINGS["settings"]["auto_sync"] = False
        self.assertEqual(auto_sync(obj), False)
        TEST_SETTINGS["settings"]["auto_sync"] = True
        self.assertEqual(auto_sync(obj), True)
        # Check that if a model is in never_auto_sync, then auto_sync returns false
        TEST_SETTINGS["settings"]["never_auto_sync"].append(
            "elasticsearch_django.testmodel"
        )
        self.assertEqual(auto_sync(obj), False)



# search.tests package identifier
from django.db import models

from ..models import SearchDocumentMixin, SearchDocumentManagerMixin


SEARCH_DOC = {"foo": "bar"}


class TestModelManager(SearchDocumentManagerMixin, models.Manager):
    pass


class TestModel(SearchDocumentMixin, models.Model):

    """Model class for use in tests."""

    simple_field_1 = models.IntegerField()
    simple_field_2 = models.CharField(max_length=100)
    complex_field = models.FileField()

    class Meta:
        # should prevent db errors during tests
        managed = False

    objects = TestModelManager()

    def as_search_document(self, index):
        return SEARCH_DOC



import warnings

warnings.warn(
    "Utils module is pending deprecation, please import disable_search_updates "
    "from decorators module instead.",
    PendingDeprecationWarning,
)

from .decorators import disable_search_updates



from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("elasticsearch_django", "0002_searchquery_duration")]

    operations = [
        migrations.AlterModelOptions(
            name="searchquery",
            options={
                "verbose_name": "Search query",
                "verbose_name_plural": "Search queries",
            },
        )
    ]



from django.db import models, migrations
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [migrations.swappable_dependency(settings.AUTH_USER_MODEL)]

    operations = [
        migrations.CreateModel(
            name="SearchQuery",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                (
                    "index",
                    models.CharField(
                        default="_all",
                        help_text="The name of the ElasticSearch index(es) being queried.",
                        max_length=100,
                    ),
                ),
                (
                    "query",
                    models.TextField(
                        default="{}", help_text="The raw ElasticSearch DSL query."
                    ),
                ),
                (
                    "hits",
                    models.TextField(
                        default="{}",
                        help_text="The list of meta info for each of the query matches returned.",
                    ),
                ),
                (
                    "total_hits",
                    models.IntegerField(
                        default=0,
                        help_text="Total number of matches found for the query (!= the hits returned).",
                    ),
                ),
                (
                    "reference",
                    models.CharField(
                        default="",
                        help_text="Custom reference used to identify and group related searches.",
                        max_length=100,
                        blank=True,
                    ),
                ),
                (
                    "executed_at",
                    models.DateTimeField(
                        help_text="When the search was executed - set via execute() method."
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        related_name="search_queries",
                        blank=True,
                        to=settings.AUTH_USER_MODEL,
                        help_text="The user who made the search query (nullable).",
                        null=True,
                        on_delete=models.SET_NULL,
                    ),
                ),
            ],
        )
    ]



from unittest import mock

from django.test import TestCase

from ..index import (
    create_index,
    update_index,
    delete_index,
    prune_index,
    _prune_hit,
    scan_index,
    bulk_actions,
)
from ..tests import TestModel, TestModelManager


class IndexFunctionTests(TestCase):

    """Test index functions."""

    @mock.patch("elasticsearch_django.index.get_client")
    @mock.patch("elasticsearch_django.index.get_index_mapping")
    def test_create_index(self, mock_mapping, mock_client):
        """Test the create_index function."""
        mock_client.return_value = mock.Mock()
        create_index("foo")
        mock_client.assert_called_once_with()
        mock_mapping.assert_called_once_with("foo")
        mock_client.return_value.indices.create.assert_called_once_with(
            index="foo", body=mock_mapping.return_value
        )

    from django.db.models.query import QuerySet

    @mock.patch.object(QuerySet, "iterator")
    @mock.patch("elasticsearch_django.index.get_client")
    @mock.patch("elasticsearch_django.index.bulk_actions")
    @mock.patch("elasticsearch_django.index.get_index_models")
    @mock.patch("elasticsearch.helpers.bulk")
    def test_update_index(
        self, mock_bulk, mock_models, mock_actions, mock_client, mock_qs
    ):
        """Test the update_index function."""
        mock_foo = mock.Mock()
        mock_foo.search_doc_type = mock.PropertyMock(return_value="bar")
        mock_foo.objects = mock.PropertyMock(return_value=mock.Mock())
        mock_models.return_value = [mock_foo]
        responses = update_index("foo")
        self.assertEqual(responses, [mock_bulk.return_value])

    @mock.patch("elasticsearch_django.index.get_client")
    def test_delete_index(self, mock_client):
        """Test the delete_index function."""
        delete_index("foo")
        mock_client.assert_called_once()
        mock_client.return_value.indices.delete.assert_called_once_with(index="foo")

    @mock.patch("elasticsearch_django.index.helpers")
    @mock.patch("elasticsearch_django.index.scan_index")
    @mock.patch("elasticsearch_django.index._prune_hit")
    @mock.patch("elasticsearch_django.index.bulk_actions")
    @mock.patch("elasticsearch_django.index.get_index_models")
    @mock.patch("elasticsearch_django.index.get_client")
    def test_prune_index(
        self,
        mock_client,
        mock_models,
        mock_actions,
        mock_prune,
        mock_scan,
        mock_helpers,
    ):
        """Test the prune_index function."""

        # this forces one single evaluation of the outer and inner for loop
        mock_models.return_value = [TestModel]
        mock_scan.return_value = ["hit"]

        # _prune_hit returns an object, so bulk should be called
        mock_prune.return_value = TestModel()
        # should return a list with one item in it
        self.assertEqual(prune_index("foo"), [mock_helpers.bulk.return_value])
        # should have called actions and bulk once each
        mock_actions.assert_called_once()
        mock_helpers.bulk.assert_called_once()

        mock_actions.reset_mock()
        mock_helpers.bulk.reset_mock()
        # if there are no objects to prune
        mock_prune.return_value = None
        # should return an empty list
        self.assertEqual(prune_index("foo"), [])
        # shouldn't call either actions or bulk (as there's no need)
        mock_actions.assert_not_called()
        mock_helpers.bulk.assert_not_called()

    @mock.patch.object(TestModelManager, "in_search_queryset")
    def test__prune_hit(self, mock_qs):
        """Test the _prune_hit function."""
        hit = {"_id": 1, "_index": "foo"}
        mock_qs.return_value = True
        self.assertIsNone(_prune_hit(hit, TestModel))

        mock_qs.return_value = False
        # should now return an instance of TestModel
        obj = _prune_hit(hit, TestModel)
        self.assertIsInstance(obj, TestModel)
        self.assertEqual(obj.id, hit["_id"])

    @mock.patch("elasticsearch_django.index.get_client")
    @mock.patch("elasticsearch_django.index.helpers")
    def test_scan_index(self, mock_helpers, mock_client):
        """Test the scan_index function."""
        query = {"query": {"type": {"value": "testmodel"}}}
        # mock_helpers.scan.return_value = ['foo', 'bar']
        # cast to list to force evaluation of the generator
        response = list(scan_index("foo", TestModel))
        mock_helpers.scan.assert_called_once_with(
            mock_client.return_value, query=query, index="foo"
        )
        self.assertEqual(response, list(mock_helpers.scan.return_value))

    @mock.patch.object(TestModel, "as_search_action")
    def test_bulk_actions(self, mock_action):
        """Test the bulk_actions function."""
        # cannot pass in in '_all' as the bulk_actions
        with self.assertRaises(AssertionError):
            list(bulk_actions([], "_all", "index"))

        mock_action.return_value = "foo"
        objects = [TestModel(), TestModel()]

        self.assertEqual(list(bulk_actions(objects, "foo", "update")), ["foo", "foo"])

        # now let's add in a bad object, and check we still get the good one
        self.assertEqual(
            list(bulk_actions([TestModel(), "bad"], "foo", "update")), ["foo"]
        )



import logging

from django.apps import AppConfig
from django.core.exceptions import ImproperlyConfigured
from django.db.models import signals

from . import settings

logger = logging.getLogger(__name__)


class ElasticAppConfig(AppConfig):

    """AppConfig for Search3."""

    name = "elasticsearch_django"
    verbose_name = "Elasticsearch"
    configs = []

    def ready(self):
        """Validate config and connect signals."""
        super(ElasticAppConfig, self).ready()
        _validate_config(settings.get_setting("strict_validation"))
        if settings.get_setting("auto_sync"):
            _connect_signals()
        else:
            logger.debug("SEARCH_AUTO_SYNC has been disabled.")


def _validate_config(strict=False):
    """Validate settings.SEARCH_SETTINGS."""
    for index in settings.get_index_names():
        _validate_mapping(index, strict=strict)
        for model in settings.get_index_models(index):
            _validate_model(model)
    if settings.get_setting("update_strategy", "full") not in ["full", "partial"]:
        raise ImproperlyConfigured(
            "Invalid SEARCH_SETTINGS: 'update_strategy' value must be 'full' or 'partial'."
        )


def _validate_mapping(index, strict=False):
    """Check that an index mapping JSON file exists."""
    try:
        settings.get_index_mapping(index)
    except IOError:
        if strict:
            raise ImproperlyConfigured("Index '%s' has no mapping file." % index)
        else:
            logger.warning("Index '%s' has no mapping, relying on ES instead.", index)


def _validate_model(model):
    """Check that a model configured for an index subclasses the required classes."""
    if not hasattr(model, "as_search_document"):
        raise ImproperlyConfigured("'%s' must implement `as_search_document`." % model)
    if not hasattr(model.objects, "get_search_queryset"):
        raise ImproperlyConfigured(
            "'%s.objects must implement `get_search_queryset`." % model
        )


def _connect_signals():
    """Connect up post_save, post_delete signals for models."""
    for index in settings.get_index_names():
        for model in settings.get_index_models(index):
            _connect_model_signals(model)


def _connect_model_signals(model):
    """Connect signals for a single model."""
    if settings.auto_sync(model):
        dispatch_uid = "%s.post_save" % model._meta.model_name
        logger.debug("Connecting search index model sync signal: %s", dispatch_uid)
        signals.post_save.connect(
            _on_model_save, sender=model, dispatch_uid=dispatch_uid
        )
        dispatch_uid = "%s.post_delete" % model._meta.model_name
        logger.debug("Connecting search index model sync signal: %s", dispatch_uid)
        signals.post_delete.connect(
            _on_model_delete, sender=model, dispatch_uid=dispatch_uid
        )
    else:
        logger.debug(
            "Auto sync disabled for '%s', ignoring update.", model._meta.model_name
        )


def _on_model_save(sender, **kwargs):
    """Update document in search index post_save."""
    update_fields = kwargs["update_fields"]
    instance = kwargs["instance"]
    for index in instance.search_indexes:
        _update_search_index(
            instance=instance, index=index, update_fields=update_fields
        )


def _update_search_index(*, instance, index, update_fields):
    """Process index / update search index update actions."""
    try:
        if update_fields:
            instance.update_search_document(index=index, update_fields=update_fields)
        else:
            instance.index_search_document(index=index)
    except Exception:
        logger.exception("Error handling 'post_save' signal for %s", instance)


def _on_model_delete(sender, **kwargs):
    """
    Remove documents from search index post_delete.

    When deleting a document from the search index, always use force=True
    to ignore the in_search_queryset check - as by definition on a delete
    the object will no longer in exist in the database.

    """
    instance = kwargs["instance"]
    for index in instance.search_indexes:
        try:
            instance.delete_search_document(index=index)
        except Exception:
            logger.exception("Error handling 'on_delete' signal for %s", instance)



"""Base command for search-related management commands."""
import builtins
import logging

from django.core.management.base import BaseCommand
from elasticsearch.exceptions import TransportError

logger = logging.getLogger(__name__)


class BaseSearchCommand(BaseCommand):

    """Base class for commands that interact with the search index."""

    description = "Base search command."

    def _confirm_action(self):
        """Return True if the user confirms the action."""
        msg = "Are you sure you wish to continue? [y/N] "
        return builtins.input(msg).lower().startswith("y")

    def add_arguments(self, parser):
        """Add default base options of --noinput and indexes."""
        parser.add_argument(
            "--noinput",
            action="store_false",
            dest="interactive",
            default=True,
            help="Do no display user prompts - may affect data.",
        )
        parser.add_argument(
            "indexes", nargs="*", help="Names of indexes on which to run the command."
        )

    def do_index_command(self, index, interactive):
        """Run a command against a named index."""
        raise NotImplementedError()

    def handle(self, *args, **options):
        """Run do_index_command on each specified index and log the output."""
        for index in options.pop("indexes"):
            data = {}
            try:
                data = self.do_index_command(index, **options)
            except TransportError as ex:
                logger.warning("ElasticSearch threw an error: %s", ex)
                data = {"index": index, "status": ex.status_code, "reason": ex.error}
            finally:
                logger.info(data)
